/*
 * lesslog - PostgreSQL archive log size reduction utility
 *
 *    Copyright(C) 2008 Nippon Telegraph and Telephone Corporation
 *
 *    Contact: suzuki.koichi@oss.ntt.co.jp
 */

/*
 * file pg_decompresslog.c
 *    Implementation of the archive restore command (pg_decompresslog).
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "postgres.h"
#include "access/xlog.h"
#include "access/xlog_internal.h"
#include "catalog/pg_control.h"

#include "file.h"
#include "remove.h"
#include "debug.h"

/* =============================================================================
 * Global variables
 * ===========================================================================*/

/* Buffer to hold restored WAL segment file image. */
static char xlog_buff[XLogSegSize];
/* Buffer to read an archive log file. */
static char arch_buff[XLogSegSize];
/* Position to write a record data. */
static char *write_pos = xlog_buff;
/* Position to read record data in the archive log buffer. */
static char *read_pos = arch_buff;
/* This holds data in the first page header of the segment. */
static XLogPageHeaderData baseheader;

/* =============================================================================
 * Prototype declaration
 * ===========================================================================*/
static void print_usage(int code);
int create_wal_image(int arch_len);
static int write_record(char *record_buff, int rem_len, bool isFromPrevSeg);
static int get_freespace(void);
static void insert_XLogContRecord(char *write_pos, int rem_len);
static void insert_pageheader(char *write_pos, XLogPageHeader pheader,
	bool hasContRecord);
static int open_arch_file(int argc, char *argv[]);
static int open_xlog_file(int argc, char *argv[]);

/* =============================================================================
 * Macro
 * ===========================================================================*/
#define IS_REMOVED_REC(record) \
	((record)->xl_info & XLR_BKP_REMOVABLE && \
	!((record)->xl_info & XLR_BKP_BLOCK_MASK))

/* =============================================================================
 * Function definitions
 * ===========================================================================*/

/*
 * Entry point of pg_decompresslog command.
 */
int
main(int argc, char *argv[])
{
	int from_fd = -1;
	int to_fd = -1;
	size_t arch_len;

	/* Error if argument(s) other than <from>, and <to> are given.  */
	if (argc > 3)
		print_usage(1);

	/* Open the archive log file to restore. */
	from_fd = open_arch_file(argc, argv);

	/*
	 * Read all the data in the input archive log file.
	 * If the header at the first page is not valid, it is not a WAL
	 * segment file and then copy the whole input file to the output
	 * file. pg_compresslog changes xlp_xlog_blcksz from XLOG_BLCKSZ
	 * to 0, so we check it to prevent from parsing normal WAL segment
	 * file as compressed file.
	 */
	arch_len = read_buff(from_fd, arch_buff, XLogSegSize);
	if (!IS_WAL_FILE(arch_buff, 0))
	{
		/*
		 * Write what is read for header validation check and then
		 * copy the rest of the input file.
		 */
		to_fd = open_xlog_file(argc, argv);
		write_buff(to_fd, arch_buff, arch_len);
		copy_file(from_fd, to_fd);
		if (close(from_fd) < 0)
		{
			fprintf(stderr, "failed to close `%s': %s\n", argv[1],
				strerror(errno));
			exit(1);
		}
		exit(0);
	}
	if (close(from_fd) < 0)
	{
		fprintf(stderr, "failed to close `%s': %s\n", argv[1], strerror(errno));
		exit(1);
	}

	/*
	 * Build the restored WAL segment file image.
	 * Write all the restored WAL segment file image.
	 */
	if (create_wal_image((int)arch_len))
	{
		fprintf(stderr, "failed to create the image of `%s'\n", argv[1]);
		exit(1);
	}
	to_fd = open_xlog_file(argc, argv);
	write_buff(to_fd, xlog_buff, XLogSegSize);
	if (close(to_fd) < 0)
	{
		fprintf(stderr, "failed to close `%s': %s\n", argv[2], strerror(errno));
		exit(1);
	}
	
	exit(0);
}

/*
 * Show the usage of the command and exit with specified code.
 */
static void
print_usage(int code)
{
	printf(
		"usage: pg_decompresslog [from [to]]\n"
		"    from - Iput file name (stdin if omitted or specified as '-')\n"
		"    to   - Output file name (stdout if omitted or specified as '-')\n"
	);
	exit(code);
}

/*
 * Restore a page of WAL segment file image from an archive log
 * build by pg_compresslog command.
 *
 * Parameters:
 *   arch_len: Size of the archive log file.
 */
int
create_wal_image(int arch_len)
{
	/* Buffer holding one record data. */
	static char record_buff[XLogSegSize];
	XLogPageHeader pheader;
	XLogContRecord *pcontrec = NULL;
	XLogRecord *pread_rec = NULL;
	XLogRecord *pwrite_rec = (XLogRecord *)record_buff;
	char *rec_write_pos = record_buff;
	bool was_last_noop = false;
	uint32 noop_len;
	XLogRecPtr removed_lsn = { 0, 0 };		/* stored in the next-record */
	uint32 removed_pos = 0;		/* initialized to avoid compile-warning */
	uint32 noop_pos = 0;		/* initialized to avoid compile-warning */

	/*
	 * Copy the archive log file page header and hold info in the
	 * header.  They are used to restore page headers of WAL segment
	 * file.  pg_compresslog changes xlp_xlog_blcksz in Page Header,
	 * so we restore it to XLOG_BLCKSZ (typically 8kB).
	 */
	pheader = (XLogPageHeader)arch_buff;
	if (XLogPageHeaderSize(pheader) != SizeOfXLogLongPHD)
	{
		fprintf(stderr, "invalid pageheader size.\n");
		return 1;
	}
	((XLogLongPageHeader)pheader)->xlp_xlog_blcksz = XLOG_BLCKSZ;
	memcpy(write_pos, (char *)pheader, SizeOfXLogLongPHD);
	read_pos += SizeOfXLogLongPHD;
	write_pos += SizeOfXLogLongPHD;

	baseheader.xlp_magic = pheader->xlp_magic;
	baseheader.xlp_info &= ~XLP_ALL_FLAGS;
	baseheader.xlp_tli = pheader->xlp_tli;
	baseheader.xlp_pageaddr.xlogid = pheader->xlp_pageaddr.xlogid;
	baseheader.xlp_pageaddr.xrecoff = pheader->xlp_pageaddr.xrecoff;

	/*
	 * Copy XLogContRecord and the continuous record to the record buffer, if
	 * there is a continuous record from the last segment file.  Then move them
	 * to WAL segment file image buffer.
	 */
	if (pheader->xlp_info & XLP_FIRST_IS_CONTRECORD)
	{
		pcontrec = (XLogContRecord *)read_pos;

		/* If the size of the continue record is not valid, it's an error. */
		if (pcontrec->xl_rem_len == 0)
		{
			fprintf(stderr,
				"invalid continue record length : xl_rem_len = %u\n",
				pcontrec->xl_rem_len);
			return 1;
		}

		/* Write first XLogContRecord to record buffer*/
		memcpy(write_pos, read_pos, SizeOfXLogContRecord);
		write_pos += SizeOfXLogContRecord;
		if (rec_write_pos + pcontrec->xl_rem_len >
				record_buff + sizeof(record_buff))
		{
			fprintf(stderr, "buffer overflow detected(first continuous data).\n");
			return 1;
		}
		memcpy(rec_write_pos, read_pos + SizeOfXLogContRecord,
			pcontrec->xl_rem_len);
		read_pos += MAXALIGN(SizeOfXLogContRecord + pcontrec->xl_rem_len);

		/* Write the continuous data to WAL segment file image buffer. */
		if (write_record(record_buff, pcontrec->xl_rem_len, true))
			return 0;
	}
	dump_page_header(((write_pos - xlog_buff) / XLOG_BLCKSZ), 
												(XLogPageHeader)xlog_buff);

	/*
	 * Loop record by record, and build each record image in the record buffer.
	 */
	while ((read_pos - arch_buff) < arch_len)
	{

		/* Set the write position of the record data. */
		pread_rec = (XLogRecord *)read_pos;

		/* 
		 * If the record data fits in the current segment, validate
		 * the record. If WAL record continues to the next segment, we
		 * cannot calculate CRC for the whole record and skip the
		 * validation. 
		 */
		if ((char *)pread_rec - arch_buff + pread_rec->xl_tot_len <= arch_len)
			if (!is_valid_record(pread_rec))
				return 1;

		/*
		 * Record other than log switch must have corresponding
		 * logical data. Refer to the comment around the line 3056 in
		 * src/backend/access/transam/xlog.c (8.2.0).
		 */
		if (IS_XLOG_SWITCH(pread_rec))
		{
			if (pread_rec->xl_len != 0)
			{
				fprintf(stderr, "invalid xlog switch record.\n");
				return 1;
			}
		}
		else if (pread_rec->xl_len == 0)
		{
			fprintf(stderr, "invalid record length.\n");
			return 1;
		}

		/* Copy whole of the record to the record buffer.  */
		if (rec_write_pos + pread_rec->xl_tot_len >
				record_buff + sizeof(record_buff))
		{
			fprintf(stderr, "buffer overflow detected(record data)\n");
			return 1;
		}
		memcpy(rec_write_pos, read_pos, pread_rec->xl_tot_len);
		read_pos += MAXALIGN(pread_rec->xl_tot_len);

		/*
		 * If the record we wrote last was a NOOP record, we must
		 * adjust xl_prev of this record to point the NOOP record.
		 * When was_last_noop is true, removed_pos and noop_pos shall
		 * be set. 
		 */
		if (was_last_noop)
			pwrite_rec->xl_prev.xrecoff = removed_lsn.xrecoff +
				(noop_pos - removed_pos);

		/*	
		 * If the record is a removed-record, get NOOP record size
		 * from tail of logical log and remove it.
		 */
		if (IS_REMOVED_REC(pwrite_rec))
		{
			memcpy(&noop_len,
				record_buff + pwrite_rec->xl_tot_len - sizeof(uint32),
				sizeof(uint32));
			pwrite_rec->xl_len -= sizeof(uint32);
			pwrite_rec->xl_tot_len -= sizeof(uint32);

			/* remember to adjust xl_prev of the next-record */
			removed_pos = write_pos - xlog_buff;
		}

		/* Update CRC if record header was modified. */
		if (was_last_noop || IS_REMOVED_REC(pwrite_rec))
			pwrite_rec->xl_crc = calc_crc(pwrite_rec);

		/*
		 * Write a record image to the WAL segment image buffer. Hey,
		 * the page size is again XLOG_BLCKSZ (typically 8kB) as the
		 * original WAL.
		 */
		if (write_record(record_buff, pwrite_rec->xl_tot_len, false))
			break;

		/* If the record is removed-record, then add NOOP record. */
		was_last_noop = false;
		if (IS_REMOVED_REC(pwrite_rec))
		{
			/*
			 * LSN of the removed-record which is needed for xl_prev
			 * of the NOOP record is stored in xl_prev of the next
			 * record. 
			 */
			removed_lsn = ((XLogRecord *)read_pos)->xl_prev;

			/* Create NOOP record in record_buff */
			memset(record_buff, 0, noop_len);
			pwrite_rec->xl_prev.xlogid = removed_lsn.xlogid;
			pwrite_rec->xl_prev.xrecoff = removed_lsn.xrecoff;
			pwrite_rec->xl_xid = pwrite_rec->xl_xid;
			pwrite_rec->xl_tot_len = noop_len;
			pwrite_rec->xl_len = noop_len - SizeOfXLogRecord;
			pwrite_rec->xl_info = XLOG_NOOP;
			pwrite_rec->xl_rmid = RM_XLOG_ID;
			pwrite_rec->xl_crc = calc_crc(pwrite_rec);

			/*
			 * Remember position of noop-record to adjust xl_prev of
			 * the next-record.
			 */
			noop_pos = write_pos - xlog_buff;

			/* Copy NOOP record to xlog_buff */
			if (write_record(record_buff, noop_len, false))
				break;

			was_last_noop = true;
		}
	}

	return 0;
}

/*
 * Build a record image to fit in a page (for WAL segment image
 * buffer).
 *
 * Return value:
 *    0: One record data build complete.
 *    1: Hit the tail of the segment.
 */
static int
write_record(char *record_buff, int len, bool has_cont)
{
	int copied = 0;
	int copy_len;
	int freespace; /* Size of the free space in the page. */

	/* Dump record content (for debug) */
	if (!has_cont)
		dumpXLogRecord(&baseheader.xlp_pageaddr, write_pos - xlog_buff,
			(XLogRecord *)record_buff);

	/* Copy record with inserting page header, if needed */
	while (len - copied > 0)
	{
		int aligned_len = 0;

		/* Determine actual length to copy to current page */
		freespace = get_freespace();
		copy_len = Min(freespace, len - copied);

		/* Copy data to current page */
		if (write_pos + copy_len > xlog_buff + sizeof(xlog_buff))
		{
			fprintf(stderr, "buffer overflow detected(record data)\n");
			exit(1);
		}
		memcpy(write_pos, record_buff + copied, copy_len);

		/* Record alignment consideration */
		if (has_cont)
			aligned_len = MAXALIGN(copy_len + SizeOfXLogContRecord) -
				SizeOfXLogContRecord;
		else
			aligned_len += MAXALIGN(copy_len);
		write_pos += aligned_len;
		freespace -= aligned_len;
		copied += copy_len;

		/* If we used up this segment, return */
		if ((write_pos - xlog_buff) >= XLogSegSize ||
				sizeof(xlog_buff) - (write_pos - xlog_buff) < SizeOfXLogRecord)
			return 1;

		/* Insert page header if necessary. */
		has_cont = (len - copied > 0);
		if (freespace < SizeOfXLogRecord)
		{
			write_pos += freespace;
			if (write_pos + SizeOfXLogShortPHD > xlog_buff + sizeof(xlog_buff))
			{
				fprintf(stderr,
					"buffer overflow detected(page header).\n");
				exit(1);
			}
			insert_pageheader(write_pos, &baseheader, has_cont);
			write_pos += SizeOfXLogShortPHD;

			/* If we have data to write yet, insert XLogContRecord too. */
			if (has_cont)
			{
				if (write_pos + SizeOfXLogContRecord >
						xlog_buff + sizeof(xlog_buff))
				{
					fprintf(stderr,
						"buffer overflow detected(XLogContRecord).\n");
					exit(1);
				}
				insert_XLogContRecord(write_pos, len - copied);
				write_pos += SizeOfXLogContRecord;
			}
		}
	}

	return 0;
}

/*
 * Calculate free space size of the page.
 */
static int
get_freespace(void)
{
	return XLOG_BLCKSZ - (write_pos - xlog_buff) % XLOG_BLCKSZ;
}

/*
 * Insert a XLogContRecord to the buffer.
 *
 * Parameters:
 *    write_pos: Write position in the buffer.
 *    rem_len: Length of the remaining record which continues from the
 *    last page.
 */
static void
insert_XLogContRecord(char *write_pos, int rem_len)
{
	XLogContRecord contrec;

	contrec.xl_rem_len = rem_len;
	memcpy(write_pos, (char *)&contrec, SizeOfXLogContRecord);
}

/*
 * Insert a page header to the buffer.
 *
 * Parameters:
 *    write_pos: Write position in the buffer.
 *    pheader: Pointer to the structure holding header info at the
 *    firt page of the segment.
 *    hasContRecord: Flag to indicate a continuous record from the
 *    last page. 
 */
static void
insert_pageheader(char *write_pos, XLogPageHeader pheader, bool hasContRecord)
{
	/*
	 * Each page header is restored using the page header at the first
	 * page of the WAL segment.   Magic number (xlp_magic), timeline
	 * id (xlp_tli) and XLOGID (xlogid) should no change within a
	 * segment and they are copied from the first page header.
	 * Continuous data (xlp_info) depends on the record of a given page.
	 * xrecoff is calculated by adding XLOG_BLCKSZ to xrecoff value in
	 * the first page header.
	 */
	pheader->xlp_info &= ~XLP_ALL_FLAGS;
	if (hasContRecord)
		pheader->xlp_info |= XLP_FIRST_IS_CONTRECORD;
	pheader->xlp_pageaddr.xrecoff += XLOG_BLCKSZ;
	memcpy(write_pos, (char *)pheader, SizeOfXLogShortPHD);
}

/*
 * Open thie archive log file to be restored and return file descriptor.
 * 
 * The first argument of the command is an input file name.
 * If omitted or specified as "-", stdin will be used as an input file.
 *
 * Parameters:
 *    argc: This is one of the argument to the command, number of the arguments.
 *    argv: This is one of the argument to the command, a pointer arry to the
 *    argument list.
 *
 * Note: If error occurs within this function, whole command will exit here
 * using exit() and the caller will not have any chance to take care of errors.
 */
static int
open_arch_file(int argc, char *argv[])
{
	int from_fd = -1;

	if (argc > 1 && strcmp(argv[1], "-") != 0)
	{
		/* Open archive log file to restore. */
		from_fd = open(argv[1], O_RDONLY, 0);
		if (from_fd < 0)
		{
			fprintf(stderr, "failed to open `%s': %s\n", argv[1],
				strerror(errno));
			exit(1);
		}

		/* Obtain the segment ID from the file name (for record dump). */
		get_segment_id(argv[1]);
	}
	else
		from_fd = fileno(stdin);

	return from_fd;
}

/*
 * Open the WAL segment file to write the restored data and return file
 * descriptor.
 * 
 * The secomd argument to the command will be regarded as an output file.
 * If omitted or specified as "-", stdout will be used as the output file.
 *
 * Parameters:
 *    argc: This is one of the argument to the command, number of the
 *    arguments. 
 *    argv: This is one of the argument to the command, a pointer arry
 *    to the argument list.
 *
 * Note: If error occurs within this function, whole command will exit
 * here using exit() and the caller will not have any chance to take
 * care of errors. 
 */
static int
open_xlog_file(int argc, char *argv[])
{
	int to_fd = -1;

	if (argc > 2 && strcmp(argv[2], "-") != 0)
	{
		/* Open the WAL segment file */
		to_fd = open(argv[2], O_RDWR | O_CREAT | O_EXCL | PG_BINARY,
			S_IRUSR | S_IWUSR);
		if (to_fd < 0)
		{
			fprintf(stderr, "failed to open `%s': %s\n", argv[2],
				strerror(errno));
			exit(1);
		}
	}
	else
		to_fd = fileno(stdout);

	return to_fd;
}
