/*
 * lesslog - PostgreSQL archive log size reduction utility
 *
 *    Copyright(C) 2008 Nippon Telegraph and Telephone Corporation
 *
 *    Contact: suzuki.koichi@oss.ntt.co.jp
 */

/*
 * remove.c
 *    replace backup block with incremental data while WAL archiving.
 */
#include "postgres.h"
#include "access/clog.h"
#include "access/gin.h"
#include "access/gist_private.h"
#include "access/htup.h"
#include "access/multixact.h"
#include "access/nbtree.h"
#include "access/transam.h"
#include "access/xact.h"
#include "access/xlog_internal.h"
#include "catalog/pg_control.h"
#include "commands/dbcommands.h"
#include "commands/tablespace.h"
#include "storage/bufpage.h"

#include "debug.h"
#include "file.h"
#include "remove.h"
#include "remove_handler.h"

/* Buffer contains WAL segument file image */
extern char xlog_buff[XLogSegSize];

/* Check if the physical log can be removed. */
#define HAS_REMOVABLE_BKP_BLOCK(record) \
	(((record)->xl_info & XLR_BKP_BLOCK_MASK) && \
	((record)->xl_info & XLR_BKP_REMOVABLE))

/*
 * Remove backup blocks in xlog record those are marked as `REMOVABLE'.
 *
 * When removing backup blocks of xlog, it needs to extract
 * incremental-format data from backup blocks.  Because archive
 * recovery needs the incremental-format data to redo changes before
 * crash, which in some cases has been optimized away because the
 * full-page-write block is present, since the block already contains
 * the changes.  If backup blocks are removed, incremental-format data
 * needs to be extracted back the optimized-away incremental-format
 * data, otherwise you won't be able to do archive recovery correctly.
 *
 * We also need to maintain the consistent of LSN in the following
 * record when we remove backup blocks. After backup blocs have been
 * removed from WAL file, the LSN information of its following xlog
 * records becomes inconsistent. The length of removed space is
 * appended to XLOG record as a hint to maintain LSN consistent during
 * recovery.
 */
uint32
remove_bkp_block(XLogRecord *record, const char *rec_pos)
{
	uint8 info = record->xl_info & ~XLR_INFO_MASK;
	uint32 noop_len = 0;
	const char *next_pos;
#ifdef DEBUG
	XLogRecord org_rec = *record;
#endif

	/* If there is no removable backup block, then do nothing. */
	if (!HAS_REMOVABLE_BKP_BLOCK(record))
		return 0;

	/*
	 * If the next record does not fit to the same segment,there's
	 * nothing we can do. 
	 */
	next_pos = get_next_rec_pos(rec_pos + record->xl_tot_len +
		get_paging_overhead(rec_pos - xlog_buff, record->xl_tot_len));
	if (!next_pos)
		return 0;

	/*
	 * Remove backup blocks from XLOG record which contains backup
	 * blocks.
	 *
	 * For each WAL record, backup block removal function is called
	 * according to the resource manager and (optionally) info field
	 * which specify record format.
	 *
	 * The following code doesn't take care XLOG records without
	 * backup blocks.  Such XLOG records should be filtered out by
	 * testing through HAS_REMOVABLE_BKP_BLOCK().
	 *
	 */
	if (record->xl_rmid == RM_HEAP_ID)
	{
		info &= XLOG_HEAP_OPMASK;

		if (info == XLOG_HEAP_INSERT)
			noop_len = remove_heap_insert(record, rec_pos, next_pos);
		else if (info == XLOG_HEAP_DELETE)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_HEAP_UPDATE)
			noop_len = remove_heap_update(record, rec_pos, next_pos, false);
		else if (info == XLOG_HEAP_MOVE)
			noop_len = remove_heap_update(record, rec_pos, next_pos, true);
		else if (info == XLOG_HEAP_HOT_UPDATE)
			noop_len = remove_heap_update(record, rec_pos, next_pos, false);
		else if (info == XLOG_HEAP_LOCK)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_HEAP_INPLACE)
			noop_len = remove_heap_inplace(record, rec_pos, next_pos);
	}
	else if (record->xl_rmid == RM_HEAP2_ID)
	{
		info &= XLOG_HEAP_OPMASK;

		if (info == XLOG_HEAP2_FREEZE)
			noop_len = remove_heap2_freeze(record, rec_pos, next_pos);
		else if (info == XLOG_HEAP2_CLEAN)
			noop_len = remove_heap2_clean(record, rec_pos, next_pos);
		else if (info == XLOG_HEAP2_CLEAN_MOVE)
			/*
			 * We don't have enough information to extract
			 * incremental-format data for
			 * XLOG_HEAP2_CLEAN_MOVE. Because this WAL record doesn't
			 * contain the information of removed heap tuples, we
			 * cannot extract incremental log.  Therefore, we don't
			 * remove the backup blocks for this type of WAL record.
			 */
			return 0;
	}
	else if (record->xl_rmid == RM_BTREE_ID)
	{
		if (info == XLOG_BTREE_INSERT_LEAF ||
			info == XLOG_BTREE_INSERT_UPPER ||
			info == XLOG_BTREE_INSERT_META)
			noop_len = remove_btree_insert(record, rec_pos, next_pos);
		else if (info == XLOG_BTREE_SPLIT_L)
			noop_len = remove_btree_split(record, rec_pos, next_pos, true);
		else if (info == XLOG_BTREE_SPLIT_R)
			noop_len = remove_btree_split(record, rec_pos, next_pos, false);
		else if (info == XLOG_BTREE_SPLIT_L_ROOT)
			noop_len = remove_btree_split(record, rec_pos, next_pos, true);
		else if (info == XLOG_BTREE_SPLIT_R_ROOT)
			noop_len = remove_btree_split(record, rec_pos, next_pos, false);
		else if (info == XLOG_BTREE_DELETE)
			/*
			 * We don't have enough information to extract
			 * incremental-format data for XLOG_BTREE_DELETE, because
			 * we don't have the information of VACUUMed heap
			 * tuples. We don't remove the backup blocks.
			 */
			return 0;
		else if (info == XLOG_BTREE_DELETE_PAGE_HALF)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_BTREE_DELETE_PAGE_META)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_BTREE_DELETE_PAGE)
			noop_len = remove_common(record, rec_pos, next_pos);
	}
	else if (record->xl_rmid == RM_GIN_ID)
	{
		if (info == XLOG_GIN_INSERT)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_GIN_VACUUM_PAGE)
			noop_len = remove_common(record, rec_pos, next_pos);
		else if (info == XLOG_GIN_DELETE_PAGE)
			noop_len = remove_common(record, rec_pos, next_pos);
	}
	else if (record->xl_rmid == RM_GIST_ID)
	{
		if (info == XLOG_GIST_PAGE_UPDATE)
			/*
			 * We don't have enough information to extract
			 * incremental-format data for XLOG_GIST_PAGE_UPDATE,
			 * because todelete array is lost when backup block
			 * presents. We don't remove the backup blocks. 
			 *
			 * To make it possible to extract incremental log
			 * resource manager should call XLogInsert() so that
			 * todelete information can survive when backup blocks are
			 * added.  This can be done by setting up formUpdateRdata as
			 *           rdata[1].buffer = InvalidBuffer;
			 * Because buffer number is set to another rdata[]
			 * element, backup blocks will be added to the WAL record.
			 */
			return 0;
		else if (info == XLOG_GIST_NEW_ROOT)
			noop_len = remove_gist_new_root(record, rec_pos, next_pos);
		else if (info == XLOG_GIST_PAGE_DELETE)
			noop_len = remove_common(record, rec_pos, next_pos);
	}

	/*
	 * Append the length of removed space to the end of XLOG record,
	 * it is used to create XNOOP record in decompress process.
	 */
	if (noop_len)
	{
		memcpy(((char *)record) + SizeOfXLogRecord + record->xl_len, &noop_len,
		sizeof(uint32));
		record->xl_info &= ~XLR_BKP_BLOCK_MASK;
		record->xl_len += sizeof(uint32);
		record->xl_tot_len = SizeOfXLogRecord + record->xl_len;
		record->xl_crc = calc_crc(record);
#ifdef DEBUG
		fprintf(stderr, "%s(): physical log removed.\n", __func__);
		fprintf(stderr, "    xl_info = 0x%02X -> 0x%02X\n", org_rec.xl_info, record->xl_info);
		fprintf(stderr, "    xl_len = %u -> %u\n", org_rec.xl_len, record->xl_len);
		fprintf(stderr, "    xl_tot_len = %u -> %u\n", org_rec.xl_tot_len, record->xl_tot_len);
#endif
	}

	return noop_len;
}

/*
 * Calculate the size of the space from `offset' to the end of the
 * page.  If the offset is on a page border, return 0 (no space left).
 */
uint32
get_space_in_page(uint32 offset)
{
	if (offset % XLOG_BLCKSZ == 0)
		return 0;
	return XLOG_BLCKSZ - (offset % XLOG_BLCKSZ);
}

/*
 * Calculate the paging overhead of an XLOG record when the XLOG record
 * spans across mutiple blocks.
 *
 * If an XLOG record doesn't fit into single block, it has to be split
 * and placed to span two or more blocks.  The overhead consists of
 * XLogContRecord and the page header. The formula to calculate the paging
 * over head is:
 *
 *   paging overhead = (SizeOfXLogShortPHD + XLogContRecord) * n
 *
 * Where n is the additional number of blocks needed to place the
 * given XLOG (n doesn't include the first block).  
 */
uint32
get_paging_overhead(uint32 rec_off, uint32 rec_len)
{
	uint32 unused_len_in_1stblk;
	int num_corssing_blks;
	int one_blk_overhead = SizeOfXLogShortPHD + SizeOfXLogContRecord;

	/* Calculate the size of unused space in the current block. */
	unused_len_in_1stblk = get_space_in_page(rec_off);

	/* If the whole record fits into the current block, returns zero. */
	if (rec_len <= unused_len_in_1stblk)
		return 0;

	/*
	 * The record needs additional blocks to fit.  Calculate how much
	 * more blocks are needed.
	 */
	num_corssing_blks =
		((rec_len - unused_len_in_1stblk) + XLOG_BLCKSZ - one_blk_overhead - 1)
		/ (XLOG_BLCKSZ - one_blk_overhead);

	return one_blk_overhead * num_corssing_blks;
}

/*
 * Find the position of the next record.
 *
 * If the next record does not fit to the rest of the current page,
 * it should begin at the next page, following the page header.
 * In thie case, NULL will be returned.
 */
const char *
get_next_rec_pos(const char *rec_end_pos)
{
	uint32 next_rec_off;
	XLogRecord *next_rec;
	uint32 overhead;
	uint32 rest;

	/*
	 * Locate the offset of next record. If the next record does not
	 * begin at the current page, then the next record should begin at
	 * the next page, following the page header.
	 */
	next_rec_off = MAXALIGN(rec_end_pos - xlog_buff);
	rest = get_space_in_page(next_rec_off);
	if (rest < SizeOfXLogRecord)
		next_rec_off += rest + SizeOfXLogShortPHD;

	/*
	 * If the next record should be at the beginning of the next WAL
	 * segment, return NULL.
	 */
	if (next_rec_off >= XLogSegSize)
		return NULL;

	/* Here, we check if the next record fits to the rest of the WAL
	 * segment file.  If it doesn't, then it should begin at the next
	 * WAL segment file.  In this case, NULL will be returned.
	 */
	next_rec = (XLogRecord *)(xlog_buff + next_rec_off);
	overhead = get_paging_overhead(next_rec_off, next_rec->xl_tot_len);
	if (next_rec_off + next_rec->xl_tot_len + overhead >= XLogSegSize)
		return NULL;

	return xlog_buff + next_rec_off;
}

/*
 * Calculate the size of XNOOP record.
 *
 */
uint32
get_noop_len(const char *rec_pos, uint32 org_len, uint32 new_len, const char *next_pos)
{
	uint32 overhead;
	uint32 removed_len_in_1stblk;
	/* offset of the XNOOP record from the head of xlog_buff. */
	uint32 del_off;
	/* size of removed area; no padding included. */
	uint32 del_size;

	overhead = get_paging_overhead(rec_pos - xlog_buff, new_len);
	del_off = MAXALIGN(rec_pos - xlog_buff + new_len + overhead);
	del_size = (rec_pos - xlog_buff) + org_len - del_off + overhead;

	/* Compute the length of removed space in first block. */
	removed_len_in_1stblk = Min(
		next_pos - xlog_buff,
		(del_off / XLOG_BLCKSZ + (del_off % XLOG_BLCKSZ ? 1 : 0)) * XLOG_BLCKSZ
	) - del_off;

	/*
	 * If the header of XNOOP record fits in current block, return the actual removed
	 * size.
	 */
	if (removed_len_in_1stblk >= SizeOfXLogRecord)
		return del_size;

	/*
	 * if XNOOP record  can be put into the following block, return
	 * ( actual removed size - padding on current block ).
	 */
	if (del_size - removed_len_in_1stblk + SizeOfXLogContRecord >= SizeOfXLogRecord)
		return del_size - removed_len_in_1stblk + SizeOfXLogContRecord;

	/*
	 * There is no enough space to place an XNOOP record, so return 0.
	 *
	 * This means that resultant incremental log and XNOOP record does
	 * not fit to the original WAL record size.  We cannot
	 * successfully replace the backup block with an incremental log
	 * and a XLOOP record.
	 */
	return 0;
}

